"""
Identifier:     csst_common/pipeline.py
Name:           file.py
Description:    pipeline operation
Author:         Bo Zhang
Created:        2023-12-13
Modified-History:
    2023-07-11, Bo Zhang, created
    2023-07-11, Bo Zhang, add Pipeline class
    2023-12-10, Bo Zhang, update Pipeline
    2023-12-15, Bo Zhang, add module header
"""
import json
import os
import shutil
import subprocess
import traceback
import warnings
from typing import Callable, NamedTuple, Optional, Any

from astropy.time import Time, TimeDelta

from .ccds import CCDS
from .dfs import DFS
from .file import File
from .logger import get_logger
from .status import CsstStatus, CsstResult


# TODO: unit test in need
class Pipeline:
    """
    CSST pipeline configuration class.

    It should be used for every pipeline to initialize environment.

    Parameters
    ----------
    dir_input : str
        Input path.
    dir_output : str
        Output path.
    dir_aux : str
        Aux path.
    dfs_root : str
        DFS root path.
    ccds_root : str
        CCDS root path.
    ccds_cache : str
        CCDS cache path.
    filter_warnings : bool
        If `True`, filter warnings.
    dfs : bool
        If `True`, initialize DFS.
    ccds : bool
        If `True`, initialize CCDS.
    **kwargs : Any
        Additional keyword arguments.
    """

    def __init__(
        self,
        dir_input: str = "/pipeline/input",
        dir_output: str = "/pipeline/output",
        dir_aux: str = "/pipeline/aux",
        dfs_root: str = "/dfs_root",
        ccds_root: str = "/ccds_root",
        ccds_cache: str = "/pipeline/ccds_cache",
        pipeline_log: str = "pipeline.log",
        module_log: str = "module.log",
        filter_warnings: bool = False,
        dfs: bool = True,
        ccds: bool = False,
        **kwargs: Any,
    ):
        # get pipeline information from env vars
        self.pipeline_id = os.getenv("PIPELINE_ID", "-")
        self.build = os.getenv("BUILD", "-")
        self.created = os.getenv("CREATED", "-")

        # set directory information
        self.dir_input = dir_input
        self.dir_output = dir_output
        self.dir_aux = dir_aux
        self.dfs_root = dfs_root
        self.ccds_root = ccds_root
        self.ccds_cache = ccds_cache

        # additional parameters
        self.kwargs = kwargs

        # set logger
        self.pipeline_logger = get_logger(
            name="pipeline logger",
            filename=os.path.join(self.dir_output, pipeline_log),
        )
        self.module_logger = get_logger(
            name="module logger",
            filename=os.path.join(self.dir_output, module_log),
        )

        # change working directory
        print(f"Change directory to {self.dir_output}")
        os.chdir(self.dir_output)

        # Frequently used files
        self.message = Message(os.path.join(self.dir_output, "message.txt"))
        self.timestamp = Timestamp(os.path.join(self.dir_output, "timestamp.txt"))
        # self.exit_code = ExitCode(os.path.join(self.dir_output, "exit_code"))
        # self.error_trace = ErrorTrace(os.path.join(self.dir_output, "error_trace"))

        if dfs:
            self.dfs = DFS(n_try=5)
        else:
            self.dfs = None
        if ccds:
            self.ccds = CCDS(ccds_root=ccds_root, ccds_cache=ccds_cache)
        else:
            self.ccds = None

        if filter_warnings:
            self.filter_warnings()

    def inspect(self):
        """List environment variables such as `PIPELINE_ID`, etc."""
        print(f"PIPELINE_ID={self.pipeline_id}")
        print(f"BUILD={self.build}")
        print(f"CREATED={self.created}")

    def clean_output(self):
        """Clean output directory."""
        self.clean_directory(self.dir_output)

    @staticmethod
    def clean_directory(d):
        """Clean a directory."""
        print(f"Clean output directory '{d}'...")
        r = subprocess.run(f"rm -rf {d}/*", shell=True, capture_output=True)
        print("> ", r)
        r.check_returncode()

    def now(self):
        """Return ISOT format datetime using `astropy`."""
        return Time.now().isot

    @staticmethod
    def filter_warnings():
        # Suppress all warnings
        warnings.filterwarnings("ignore")

    @staticmethod
    def reset_warnings(self):
        """Reset warning filters."""
        warnings.resetwarnings()

    def file(self, file_path):
        """Initialize File object."""
        return File(file_path, new_dir=self.dir_output)

    def new(self, file_name="test.fits"):
        """Create new file in output directory."""
        return os.path.join(self.dir_output, file_name)

    @staticmethod
    def move(file_src: str, file_dst: str):
        """Move file `file_src` to `file_dist`."""
        shutil.move(file_src, file_dst)

    @staticmethod
    def copy(file_src: str, file_dst: str):
        """Move file `file_src` to `file_dist`."""
        shutil.copy(file_src, file_dst)

    def call(self, func: Callable, *args: Any, **kwargs: Any):
        self.pipeline_logger.info(
            f"====================================================="
        )
        t_start: Time = Time.now()
        self.pipeline_logger.info(f"Starting Module: **{func.__name__}**")
        # logger.info(f"Additional arguments: {args} {kwargs}")
        try:
            # if the module works well
            res: CsstResult = func(*args, **kwargs)
            assert isinstance(res, CsstResult)
            # define results
            status = res.status
            files = res.files
            output = res.output
        except Exception as e:
            # if the module raises error
            exc_info = traceback.format_exc()  # traceback info
            self.pipeline_logger.error(f"Error occurs! \n{exc_info}")
            # define results
            status = CsstStatus.ERROR  # default status if exceptions occur
            files = None
            output = {"exc_info": exc_info}  # default output if exceptions occur
        finally:
            t_stop: Time = Time.now()
            t_cost: float = (t_stop - t_start).value
            if isinstance(status, CsstStatus):
                # status is
                self.pipeline_logger.info(
                    f"Module finished: status={status} | cost={t_cost:.1f} sec"
                )
            else:
                # invalid status
                self.pipeline_logger.error(
                    f"Invalid status: {status} is not a CsstResult object!"
                )
            # record exception traceback info
            self.pipeline_logger.info(
                f"ModuleResult: \n"
                f"   - name: {func.__name__}\n"
                f"   - status: {status}\n"
                f"   - files: {files}\n"
                f"   - output: {output}\n"
            )
            return ModuleResult(
                module=func.__name__,
                cost=t_cost,
                status=status,
                files=files,
                output=output,
            )


# class ErrorTrace:
#     """Write error trace to file."""
#
#     def __init__(self, file_path=""):
#         self.file_path = file_path
#
#     def __repr__(self):
#         return f"< ErrorTrace [{self.file_path}] >"
#
#     def write(self, s: str):
#         with open(self.file_path, "w+") as f:
#             f.write(s)


class Message:
    """Write JSON format messages to file."""

    def __init__(self, file_path: str = ""):
        self.file_path = file_path

    def __repr__(self):
        return f"< Message [{self.file_path}] >"

    def write(self, dlist):
        with open(self.file_path, "w+") as f:
            for d in dlist:
                f.write(self.dict2msg(d) + "\n")

    @staticmethod
    def dict2msg(d):
        m = json.dumps(d).replace(" ", "")
        return m

    @staticmethod
    def msg2dict(m):
        d = json.loads(m)
        return d


# DEPRECATED
# class ExitCode:
#     def __init__(self, file_path=""):
#         self.file_path = file_path
#
#     def __repr__(self):
#         return f"< ExitCode [{self.file_path}] >"
#
#     def truncate(self):
#         with open(self.file_path, "w") as file:
#             file.truncate(0)
#
#     def write(self, code=0):
#         with open(self.file_path, "w+") as f:
#             f.write(str(code))
#         print(f"Exit with code {code} (written to '{self.file_path}')")


class Timestamp:
    def __init__(self, file_path: str = "timestamp.txt"):
        """
        Timestamp Class.

        Initialize a Timestamp object anc connect it to `file_path`.

        Parameters
        ----------
        file_path : str
            Time stamp file path.
        """
        self.file_path = file_path

    def __repr__(self):
        return f"< Timestamp [{self.file_path}] >"

    def empty(self):
        """Clean time stamp file."""
        with open(self.file_path, "w") as file:
            file.truncate(0)

    def touch(self):
        """Write a time stamp."""
        with open(self.file_path, "a+") as f:
            f.write(f"{Time.now().isot}+00:00\n")


class ModuleResult(NamedTuple):
    module: str
    cost: float
    status: CsstStatus
    files: Optional[list]
    output: dict
